/*
 * $Id: zrp.h,v 1.1 2000/03/28 22:52:08 xkx Exp $
 * This is the header file for zrp routing protocol.
 * Following the specification of ZRP Internet Draft.
 * <draft-ietf-manet-zone-zrp-02.txt>
 *
 * This implementation is based on the former implementation
 * by Mark Yarvis and Joy Kongkitisupchai.
 */

#ifndef _ZRP_H_
#define _ZRP_H_

/*
 * General definition of contants used in ZRP.
 */

// maximum number of nodes in a route 
// NOTE: fix the size of RoutingZrpNetworkDU
// #define ROUTING_ZRP_MAX_ROUTE_LEN  100
#define ROUTING_ZRP_MAX_ROUTE_LEN   50

// Use to re-trigger stale IERP route query 1 SEC=1000000000
#define ROUTING_ZRP_QUERY_TIMEOUT 20 * SECOND

// maximum number of retx of IERP query if the previous was stale
#define ROUTING_ZRP_MAX_QUERY_RETX  5

// null IERP route query Id
#define NULL_QUERY_ID -1

/*
 * Definitions of contants for IARP.
 */
// Unreacheable route has hop count >= this value.
#define ROUTING_IARP_INFINITY (0x7fff)

// Used to check error conditions
#define ROUTING_IARP_INVALID_NODE    -1

/*
 * Definitions of contants for IERP.
 */
#define ROUTING_IERP_MAX_QUERY_HOPS 100

// The retry times of dropped packet to neighbors.
#define ROUTING_ZRP_PACKET_RETRY     5

/*
 * Definitions of contants for BRP
 */
// The validation time of a route query cache.
#define ROUTING_BRP_CACHE_FLUSH      ((clocktype)30 * SECOND)


/*
 * Definitions of contants for NDM
 */
#define ROUTING_ZRP_NDM_INTERVAL (30 * SECOND)
#define ROUTING_ZRP_NDM_RANDOM   10 * SECOND
#define ROUTING_ZRP_NDM_MISSTIME      3   

/*
 * General definitions of data structure for ZRP.
 */

// ZRP level statistics (not including IARP and IERP statistics)
typedef struct RoutingZrpRoutingStats_t
{
    // How many NDM message send
    int numNdmPackets;

    // How many Source Routing Failure notice packets sent.
    int numLinkFailPackets;

    // Total end-to-end delay for all data successfully sent to Trans
    // Need to div this by numToTransport for avg end-to-end delay
    clocktype totalEnd2EndDelay;

    // Total delay (for IERP query) of all data packets
    clocktype totalQueryDelay;

    // Total number of IERP query initiated
    // If query is re-initiated under the same query ID, due to
    // timeout, this re-initiated is not included in this number
    int numQueryInit;

} RoutingZrpRoutingStats;

typedef enum
{
    // IARP packet
    ROUTING_ZRP_IARP,

    // IERP packet
    ROUTING_ZRP_IERP,

    // NDM packet
    ROUTING_ZRP_NDM,

    // Route Failure packet.
    ROUTING_ZRP_LINKFAIL,

    // Data packet
    ROUTING_ZRP_DATA
} RoutingZrpPacketType;

typedef struct RoutingZrpCacheEntryNode_t{

    // time of initial transmission request
    clocktype initTime;
    NODE_ADDR destAddr;

    Message *msg;
    struct RoutingZrpCacheEntryNode_t *next;
} RoutingZrpCacheEntryNode;

// Entry to transport packet cache
typedef struct RoutingZrpCacheEntry_t {
    // unique query
    long queryId;
    NODE_ADDR destAddr;

    // when current time is rtxTime, this query is stale
    clocktype rtxTime;

    // Number of re-transmission if query is stale the previous time
    int rtxCount;

    RoutingZrpCacheEntryNode *head;
    int size;
    
    struct RoutingZrpCacheEntry_t *nextEntry;

} RoutingZrpCacheEntry;

// Use to cache packet from transport while route discovery is in progress
typedef struct {
    RoutingZrpCacheEntry *entryList;
    int size;

} RoutingZrpCache;


/*
 * Definitions of data structures for IARP.
 */

// IARP control data packet
// This packet is exchanged  among
// neighbors when route updated occurs
typedef struct RoutingIarpPacket_t {

    // type of the packet.
    RoutingZrpPacketType pktType;   // must be ROUTING_ZRP_IARP
    
    // Destination address.
    NODE_ADDR destAddr;

    // Next Hop address.
    NODE_ADDR nextHopAddr;

    // Next Next Hop address.
    NODE_ADDR nextNextHopAddr;

    // Number of hops the packet has travelled.
    short hopCount;

} RoutingIarpPacket;

// Describes one hop in routing path .
typedef struct  RoutingIarpRoutingPath_t
{
    // Node id of the next hop node.
    NODE_ADDR nextHop;

    // This distance is measured in terms of number of hops.
    short dist;
    
    struct RoutingIarpRoutingPath_t * nextPath;

} RoutingIarpRoutingPath;

// Describes one entry in routing table.
typedef struct  RoutingIarpRoutingTableRow_t
{
    // The destination node.
    NODE_ADDR destAddr;
    
    // The last time heard a message from this node.
    clocktype lastHeardTime;

    //  Retry time while packets are dropped to this dest node.
    int numRetry;

    // Number of entries in the route 
    int numRoute;

    // path towards the destID 
    RoutingIarpRoutingPath *firstRoute;
    
    struct RoutingIarpRoutingTableRow_t *nextRow;

} RoutingIarpRoutingTableRow;

// Describes routing table information for all nodes.
typedef struct RoutingIarpRoutingTable_t
{
    // Number of Rows in the routing table
    int numRow;

    // One row for each destination.
    RoutingIarpRoutingTableRow *firstRow;

} RoutingIarpRoutingTable;

// IARP statistics
typedef struct RoutingIarpRoutingStats_t
{
    // number of node lost interrupt received from NDM
    int numNodeLostIntrp;

    // number of node found interrupt received from NDM
    int numNodeFoundIntrp;

    // number of packets used to notify neighbors of updated routes
    int numRouteUpdatePkt;

} RoutingIarpRoutingStats;

// Contains info each IARP node needs to know of itself
typedef struct RoutingIarpNodeInfo_t
{
    // Statistical Info
    RoutingIarpRoutingStats stats;

    // Routing table
    RoutingIarpRoutingTable routeTable;
    
} RoutingIarpNodeInfo;


/*
 * Definitions of data structures for IERP.
 */

typedef enum
{
   ROUTING_IERP_QUERY,
   ROUTING_IERP_QUERY_EXTENSION,
   ROUTING_IERP_QUERY_REPLY,
   ROUTING_IERP_ROUTE_ACCUMULATION
/* Hook for route optimization
   ROUTING_IERP_ROUTE_OPTIMIZATION
*/
} RoutingIerpPacketType;

typedef struct RoutingIerpPacket_t {
    //zrp type of this packet.
    RoutingZrpPacketType pktType;   //must be ROUTING_ZRP_IERP
    
    RoutingIerpPacketType type;

    int hopsLeft;

    short queryId;

    NODE_ADDR destAddr;
    NODE_ADDR badLinkSourceAddr;
 
    NODE_ADDR nextIERP;
 
    NODE_ADDR nextBRP;

    NODE_ADDR prevIERP;

/* Hook for route optimization
   long flags;
*/

    NODE_ADDR route[ROUTING_ZRP_MAX_ROUTE_LEN];

    // this item is not actually transmitted
    int routeLen;

} RoutingIerpPacket;

typedef struct RoutingIerpRoute_t {
   int routeLen;
   NODE_ADDR *nodes;
   
   struct RoutingIerpRoute_t * nextRoute;
} RoutingIerpRoute;

typedef struct RoutingIerpRouteTable_Entry_t {
   NODE_ADDR dest;
   int numRoutes;
   RoutingIerpRoute *firstRoute;   
   
   struct RoutingIerpRouteTable_Entry_t * nextEntry;
} RoutingIerpRouteTableEntry;

typedef struct RoutingIerpRouteTable_t {
   int numDests;
   RoutingIerpRouteTableEntry *firstEntry;   
} RoutingIerpRouteTable;

typedef struct RoutingIerpStats_t {
   int numNodeLostIntrp;
   int msgsReceived;
   int msgsSent;
} RoutingIerpStats;

typedef struct RoutingIerpNodeInfo_t {
   long nextQuery;
   RoutingIerpRouteTable routeTable;
   RoutingIerpStats stats;
} RoutingIerpNodeInfo;


/*
 * Definitions of data structures for BRP.
 */
typedef struct RoutingBrpQueryCacheEntry_t {
   NODE_ADDR source;
   long queryId;
   clocktype deleteTime;    // This time later, we will delete this query to free memory
   NODE_ADDR prevIERP;
   NODE_ADDR prevHops[ROUTING_ZRP_MAX_ROUTE_LEN];
   int prevHopsLen;
   NODE_ADDR nextHops[ROUTING_ZRP_MAX_ROUTE_LEN];
   int nextHopsLen;
   
   struct RoutingBrpQueryCacheEntry_t *nextEntry;   
} RoutingBrpQueryCacheEntry;

typedef struct RoutingBrpStats_t {
   int msgsToMac;
   int msgsFromMac;
   int msgsToIERP;
   int msgsFromIERP;
} RoutingBrpStats;

typedef struct RoutingBrpNodeInfo_t {
   int cachedEntries;
   RoutingBrpQueryCacheEntry *detectedQueries;
   RoutingBrpStats stats;
} RoutingBrpNodeInfo;

/*
 * Data structure definitions of NDM
 */
typedef struct RoutingZrpNdmPacket_t {
    RoutingZrpPacketType pktType;  /* must be ROUTING_ZRP_NDM */
} RoutingZrpNdmPacket;
  

/*
 * Data structure for link failure notice.
 */
typedef struct RoutingZrpLinkFailPacket_t {
    RoutingZrpPacketType pktType;  /* must be ROUTING_ZRP_LINKFAIL */
    NODE_ADDR   sourceAddr;
    NODE_ADDR   destAddr;
} RoutingZrpLinkFailPacket;

// ZRP data structure 
// This is the info that each node in the ZRP knows
typedef struct glomo_routing_zrp_str {
    struct RoutingIarpNodeInfo_t iarp; 
    struct RoutingIerpNodeInfo_t ierp; 
    struct RoutingBrpNodeInfo_t brp; 

    RoutingZrpCache pktCache;
    RoutingZrpRoutingStats stats;
    int zoneRadius;
    // long layer;   Use to set transport layer to UDP to run tg_cbr
}   GlomoRoutingZrp;


/*
 * Prototype of functions for ZRP.
 */
void RoutingZrpInit(GlomoNode *nodePtr,
                    GlomoRoutingZrp **zrpPtr,
                    const GlomoNodeInput *nodeInput);

void RoutingZrpInitStats(RoutingZrpRoutingStats *zrpStats);

void RoutingZrpInitCache(RoutingZrpCache * pktCache);

void RoutingZrpFinalize(GlomoNode *nodePtr);

void RoutingZrpPrintWholeStatsInfo(GlomoNode *nodePtr, FILE *outfd);

void RoutingZrpHandleProtocolPacket(GlomoNode *nodePtr, Message *msg,
                                    NODE_ADDR srcAddr, NODE_ADDR destAddr);

void RoutingZrpHandleProtocolEvent(GlomoNode *nodePtr, Message *msg);

void RoutingZrpRouterFunction(GlomoNode * nodePtr, Message * msg,
                              NODE_ADDR destAddr, BOOL *packetWasRouted);
                              
void RoutingZrpMacLayerStatusHandler(GlomoNode *nodePtr, const Message* msg);

void RoutingZrpTransmitData(GlomoNode *nodePtr, Message *msg, NODE_ADDR nextHop,
                            NODE_ADDR *route, int routeLen);

void RoutingZrpPrepareForQueryResult(GlomoNode *nodePtr, long queryId, Message *msg,
                                     NODE_ADDR nextHop, NODE_ADDR destAddr);

void RoutingZrpCheckQueryTimeout(GlomoNode *nodePtr);

void RoutingZrpHandleIERPResult(GlomoNode *nodePtr, long queryId, 
                                NODE_ADDR *route, int routeLen);

int RoutingZrpGetIndex(NODE_ADDR target, NODE_ADDR *nodes, int len);

int RoutingZrpFindLink(NODE_ADDR sourceAddr, NODE_ADDR destAddr, NODE_ADDR *nodes, int len);

void RoutingZrpInsertNodes(NODE_ADDR *targetNodes, int targetPos,
                           int *targetLenPtr,
                           NODE_ADDR *sourceNodes, int sourceLen);

void RoutingZrpDeleteNodes(NODE_ADDR *nodes, int delPos, int delLen,
                            int *listLenPtr);

void RoutingZrpReverseNodes(NODE_ADDR *nodes, int len);

long RoutingZrpFindDestInCache(GlomoNode *nodePtr, NODE_ADDR destAddr);

void RoutingZrpPrintQueryCache(GlomoNode *nodePtr);

BOOL RoutingZrpAddToCachedPktList(GlomoNode *nodePtr, long queryId, 
                                  Message *msg, NODE_ADDR destAddr);

BOOL RoutingZrpAddToCache(GlomoNode *nodePtr, long queryId, Message *msg,
                          NODE_ADDR nextHop, NODE_ADDR destAddr);

BOOL RoutingZrpDelFromCache(GlomoNode *nodePtr, long queryId);

RoutingZrpCacheEntry *RoutingZrpFindQuery(GlomoNode *nodePtr, long queryId);

void RoutingZrpNoticeLinkFailure(GlomoNode *nodePtr, NODE_ADDR *route,
                                 int routeLen, int current);

BOOL RoutingZrpShouldRetry(GlomoNode *nodePtr, NODE_ADDR destAddr);

/*
 * Prototype of functions for NDM
 */
void RoutingZrpNdmNoticeNeighbor(GlomoNode *nodePtr);

void RoutingZrpNdmUpdateNeighbor(GlomoNode *nodePtr, NODE_ADDR newNeighAddr);

void RoutingZrpNdmSetNeighborTimeoutTimer(GlomoNode *nodePtr);

void RoutingZrpNdmCheckNeighborTimeout(GlomoNode *nodePtr);


/* 
 * Prototype of functions for IARP.
 */
void RoutingIarpInit(GlomoNode *nodePtr);

void RoutingIarpFinalize(GlomoNode *nodePtr, FILE *stat_out);

void RoutingIarpPrintRoutingTable(GlomoNode *nodePtr, 
                                  RoutingIarpRoutingTable *routeTable,
                                  BOOL toFile);

RoutingIarpRoutingTableRow * RoutingIarpFindRoutingTableRow( 
                             RoutingIarpRoutingTable * routeTable, NODE_ADDR destAddr);

void RoutingIarpInitRoutingStats(GlomoNode *nodePtr);

void RoutingIarpInitIARP(GlomoNode *nodePtr, RoutingIarpPacket *iarpPacket,
                         NODE_ADDR destAddr);

BOOL RoutingIarpFindBestRoute(GlomoNode *nodePtr, NODE_ADDR unknDestAddr,
                              RoutingIarpRoutingPath *bestRoute);

void RoutingIarpHandleProtocolPacket(GlomoNode *nodePtr, Message *msg);

void RoutingIarpHandleZoneNodeLostIntrp(GlomoNode *nodePtr, 
                                        NODE_ADDR lostNodeAddr);

void RoutingIarpHandleNeighFoundIntrp(GlomoNode *nodePtr, NODE_ADDR newNeighAddr);

BOOL RoutingIarpAdvertiseOptimalRoute(GlomoNode *nodePtr, 
                                      RoutingIarpPacket *iarpPkt,
                                      NODE_ADDR host);

void RoutingIarpUpdateRT(GlomoNode *nodePtr, NODE_ADDR host,
                         RoutingIarpPacket *iarpPkt);

BOOL RoutingIarpAddRoutingEntry(GlomoNode *nodePtr, 
                                RoutingIarpRoutingPath *newRoute,
                                NODE_ADDR newNeighAddr);

BOOL RoutingIarpDelRoutingEntry(GlomoNode *nodePtr, NODE_ADDR targetHostAddr,
                                RoutingIarpRoutingPath newRoutePath);

BOOL RoutingIarpRouteExists(GlomoNode *nodePtr, NODE_ADDR destHostAddr);

NODE_ADDR RoutingIarpNextHopTo(GlomoNode *nodePtr, NODE_ADDR destHostAddr);

int RoutingIarpDistanceToNode(GlomoNode *nodePtr, NODE_ADDR destHostAddr);


/*
 * Prototype of functions for IERP.
 */
void RoutingIerpInit(GlomoNode *nodePtr);

void RoutingIerpFinalize(GlomoNode *nodePtr, FILE *stat_out);

long RoutingIerpGetNextQueryId(GlomoNode *nodePtr);

BOOL RoutingIerpFindRoute(GlomoNode *nodePtr, NODE_ADDR dest, 
                          NODE_ADDR *route, int *routeLen);

void RoutingIerpInitiateQuery(GlomoNode *nodePtr, NODE_ADDR target, long queryId);

void RoutingIerpInitiateRepair(GlomoNode *nodePtr, NODE_ADDR *route,
                               int routeLen, int goodPortion, long queryId);

void RoutingIerpHandlePacketFromBRP(GlomoNode *nodePtr, 
                                    RoutingIerpPacket * ierpPkt);

RoutingIerpRouteTableEntry *
            RoutingIerpFindRouteTableEntry(GlomoNode *nodePtr, NODE_ADDR dest);

RoutingIerpRouteTableEntry *
            RoutingIerpMakeNewRouteTableEntry(GlomoNode *nodePtr, NODE_ADDR dest);

void RoutingIerpInsertRoute(RoutingIerpRouteTableEntry *tableEntry,
                            RoutingIerpRoute *route);

void RoutingIerpCacheRoute(GlomoNode *nodePtr, 
                           NODE_ADDR *route, int routeLen, NODE_ADDR dest);

void RoutingIerpZoneNodeLost(GlomoNode *nodePtr, NODE_ADDR lostNode);

void RoutingIerpDeleteFailedRoute(GlomoNode *nodePtr, NODE_ADDR sourceAddr, 
                                  NODE_ADDR destAddr);

/*
 * Prototype of functions for BRP.
 */
void RoutingBrpInit(GlomoNode *nodePtr);

void RoutingBrpFinalize(GlomoNode *nodePtr, FILE *stat_out);

void RoutingBrpHandleProtocolPacket(GlomoNode *nodePtr, Message *msg) ;

void RoutingBrpHandlePacketFromIERP(GlomoNode *nodePtr, 
                                    RoutingIerpPacket *ierpPkt);

void RoutingBrpForwardViaMac(GlomoNode *nodePtr, 
                             RoutingIerpPacket *ierpPkt, 
                             NODE_ADDR source, NODE_ADDR destAddr);

void RoutingBrpRemovePrevHops(RoutingIerpPacket *ierpPkt, int *localNodeIndex);

void RoutingBrpRemoveNextHops(RoutingIerpPacket *ierpPktPtr, int localNodeIndex);

BOOL RoutingBrpCacheQuery(GlomoNode *nodePtr, RoutingIerpPacket *ierpPkt,
                          int localNodeIndex, int cacheFlag);

void RoutingBrpAddLocalAddrToRoute(GlomoNode *nodePtr, 
                                   RoutingIerpPacket *ierpPkt,
                                   int *localNodeIndex);

RoutingBrpQueryCacheEntry *RoutingBrpFindQuery(GlomoNode *nodePtr, 
                                               NODE_ADDR source, long queryId);

BOOL RoutingBrpLoopExist(GlomoNode *nodePtr, NODE_ADDR *nodes, int routeLen);

void RoutingBrpRestoreNextHopsFromCache(GlomoNode *nodePtr, 
                                        RoutingIerpPacket *ierpPkt, 
                                        int localNodeIndex);

void RoutingBrpRestorePrevHopsFromCache(GlomoNode *nodePtr, 
                                        RoutingIerpPacket *ierpPkt, 
                                        int *localNodeIndex);

void RoutingBrpSetCacheFlushTimer(GlomoNode *nodePtr);

void RoutingBrpDeleteCacheEntry(GlomoNode *nodePtr);

#endif /* _ZRP_H_ */
